# MIT License
#
# Copyright (c) 2019 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Compression utilities for tskit tree sequences.
"""
import contextlib
import json
import logging
import os.path
import tempfile
import warnings
import zipfile
from typing import Mapping

import humanize
import numcodecs
import numpy as np
import tskit
import zarr

from . import exceptions
from . import provenance

logger = logging.getLogger(__name__)

FORMAT_NAME = "tszip"
FORMAT_VERSION = [1, 0]


def minimal_dtype(array):
    """
    Returns the smallest dtype that can be used to represent the values in the
    specified array. If the array is not one of the integer types, the dtype of the
    array itself is returned directly.
    """
    dtype = array.dtype
    if array.shape[0] == 0:
        return dtype
    if dtype.kind == "u":
        maxval = np.max(array)
        dtypes = [np.uint8, np.uint16, np.uint32, np.uint64]
        for dtype in map(np.dtype, dtypes):
            if maxval <= np.iinfo(dtype).max:
                break
    elif dtype.kind == "i":
        minval = np.min(array)
        maxval = np.max(array)
        dtypes = [np.int8, np.int16, np.int32, np.int64]
        for dtype in map(np.dtype, dtypes):
            if np.iinfo(dtype).min <= minval and maxval <= np.iinfo(dtype).max:
                break
    return dtype


def compress(ts, path, variants_only=False):
    """
    Compresses the specified tree sequence and writes it to the specified path.
    By default, fully lossless compression is used so that tree sequences are
    identical before and after compression. By specifying the ``variants_only``
    option, a lossy compression can be used, which discards any information
    that is not needed to represent the variants (which are stored losslessly).

    :param tskit.TreeSequence ts: The input tree sequence.
    :param str destination: The string or :class:`pathlib.Path` instance describing
        the location of the compressed file.
    :param bool variants_only: If True, discard all information not necessary
        to represent the variants in the input file.
    """
    destination = str(path)
    # Write the file into a temporary directory on the same file system so that
    # we can write the output atomically.
    destdir = os.path.dirname(os.path.abspath(destination))
    with tempfile.TemporaryDirectory(dir=destdir, prefix=".tszip_work_") as tmpdir:
        filename = os.path.join(tmpdir, "tmp.trees.tgz")
        logging.debug(f"Writing to temporary file {filename}")
        with zarr.ZipStore(filename, mode="w") as store:
            root = zarr.group(store=store)
            compress_zarr(ts, root, variants_only=variants_only)
        os.replace(filename, destination)
    logging.info(f"Wrote {destination}")


def decompress(path):
    """
    Decompresses the tszip compressed file and returns a tskit tree sequence
    instance.

    :param str path: The location of the tszip compressed file to load.
    :rtype: tskit.TreeSequence
    :return: A :class:`tskit.TreeSequence` instance corresponding to the
        the specified file.
    """
    with load_zarr(path) as root:
        return decompress_zarr(root)


class Column:
    """
    A single column that is stored in the compressed output.
    """

    def __init__(self, name, array, delta_filter=False):
        self.name = name
        self.array = array
        self.delta_filter = delta_filter

    def compress(self, root, compressor):
        shape = self.array.shape
        chunks = shape
        if shape[0] == 0:
            chunks = (1,)
        dtype = minimal_dtype(self.array)
        filters = None
        if self.delta_filter:
            filters = [numcodecs.Delta(dtype=dtype)]
        compressed_array = root.empty(
            self.name,
            chunks=chunks,
            shape=shape,
            dtype=dtype,
            filters=filters,
            compressor=compressor,
        )
        compressed_array[:] = self.array
        ratio = 0
        if compressed_array.nbytes > 0:
            ratio = compressed_array.nbytes / compressed_array.nbytes_stored
        logger.debug(
            "{}: output={} compression={:.1f}".format(
                self.name,
                humanize.naturalsize(compressed_array.nbytes_stored, binary=True),
                ratio,
            )
        )


def compress_zarr(ts, root, variants_only=False):

    provenance_dict = provenance.get_provenance_dict({"variants_only": variants_only})

    if variants_only:
        logging.info("Using lossy variants-only compression")
        # Reduce to site topology. Note that we will remove
        # any sites, individuals and populations here that have no references.
        ts = ts.simplify(reduce_to_site_topology=True)

    tables = ts.tables

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # When using a zipfile in Zarr we get some harmless warnings. See
        # https://zarr.readthedocs.io/en/stable/api/storage.html#zarr.storage.ZipStore
        root.attrs["format_name"] = FORMAT_NAME
        root.attrs["format_version"] = FORMAT_VERSION
        root.attrs["sequence_length"] = tables.sequence_length
        root.attrs["provenance"] = provenance_dict

    columns = {}
    for key, value in tables.asdict().items():
        if isinstance(value, dict):
            for sub_key, sub_value in value.items():
                columns[f"{key}/{sub_key}"] = sub_value
        else:
            columns[key] = value

    if variants_only:
        time = np.unique(tables.nodes.time)
        columns["node/time"] = np.searchsorted(time, tables.nodes.time)

    # Encoding array is a tuple so must be converted
    columns["encoding_version"] = np.asarray(columns["encoding_version"])

    # Sequence length is stored as an attr for compatibility with older versions of tszip
    del columns["sequence_length"]

    # Schemas and metadata need to be converted to arrays
    for name in columns:
        if name.endswith("metadata_schema"):
            columns[name] = np.frombuffer(columns[name].encode("utf-8"), np.int8)
        if name.endswith("metadata"):
            columns[name] = np.frombuffer(columns[name], np.int8)

    # Some columns benefit from being quantised
    coordinates = np.unique(
        np.hstack(
            [
                [0, ts.sequence_length],
                tables.edges.left,
                tables.edges.right,
                tables.sites.position,
                tables.migrations.left,
                tables.migrations.right,
            ]
        )
    )
    columns["coordinates"] = coordinates
    for name in [
        "edges/left",
        "edges/right",
        "migrations/left",
        "migrations/right",
        "sites/position",
    ]:
        columns[name] = np.searchsorted(coordinates, columns[name])

    # Some columns benefit from additional options
    delta_filter_cols = ["edges/parent", "sites/position"]

    # Note: we're not providing any options to set this here because Blosc+Zstd seems to
    # have a clear advantage in compression performance and speed. There is very little
    # difference between compression level 6 and 9, and it's extremely fast in any case
    # so there's no point in adding complexity. The shuffle filter in particular makes
    # big difference.
    compressor = numcodecs.Blosc(
        cname="zstd", clevel=9, shuffle=numcodecs.Blosc.SHUFFLE
    )
    for name, data in columns.items():
        Column(
            name, data, delta_filter="_offset" in name or name in delta_filter_cols
        ).compress(root, compressor)


def check_format(root):
    try:
        format_name = root.attrs["format_name"]
        format_version = root.attrs["format_version"]
    except KeyError as ke:
        raise exceptions.FileFormatError("Incorrect file format") from ke
    if format_name != FORMAT_NAME:
        raise exceptions.FileFormatError(
            "Incorrect file format: expected '{}' got '{}'".format(
                FORMAT_NAME, format_name
            )
        )
    if format_version[0] < FORMAT_VERSION[0]:
        raise exceptions.FileFormatError(
            "Format version {} too old. Current version = {}".format(
                format_version, FORMAT_VERSION
            )
        )
    if format_version[0] > FORMAT_VERSION[0]:
        raise exceptions.FileFormatError(
            "Format version {} too new. Current version = {}".format(
                format_version, FORMAT_VERSION
            )
        )


@contextlib.contextmanager
def load_zarr(path):
    path = str(path)
    try:
        store = zarr.ZipStore(path, mode="r")
    except zipfile.BadZipFile as bzf:
        raise exceptions.FileFormatError("File is not in tgzip format") from bzf
    root = zarr.group(store=store)
    try:
        check_format(root)
        yield root
    finally:
        store.close()


def decompress_zarr(root):
    coordinates = root["coordinates"][:]
    dict_repr = {"sequence_length": root.attrs["sequence_length"]}

    quantised_arrays = [
        "edges/left",
        "edges/right",
        "migrations/left",
        "migrations/right",
        "sites/position",
    ]
    for key, value in root.items():
        if isinstance(value, Mapping):
            for sub_key, sub_value in value.items():
                if f"{key}/{sub_key}" in quantised_arrays:
                    dict_repr.setdefault(key, {})[sub_key] = coordinates[sub_value]
                elif sub_key.endswith("metadata_schema"):
                    dict_repr.setdefault(key, {})[sub_key] = bytes(sub_value).decode(
                        "utf-8"
                    )
                else:
                    dict_repr.setdefault(key, {})[sub_key] = sub_value
        elif key.endswith("metadata_schema"):
            dict_repr[key] = bytes(value).decode("utf-8")
        elif key.endswith("metadata"):
            dict_repr[key] = bytes(value)
        else:
            dict_repr[key] = value

    return tskit.TableCollection.fromdict(dict_repr).tree_sequence()


def print_summary(path, verbosity=0):
    arrays = []

    def visitor(array):
        if isinstance(array, zarr.core.Array):
            arrays.append(array)

    with load_zarr(path) as root:
        root.visitvalues(visitor)

    arrays.sort(key=lambda x: x.nbytes_stored)
    max_name_len = max(len(array.name) for array in arrays)
    stored = [
        humanize.naturalsize(array.nbytes_stored, binary=True) for array in arrays
    ]
    max_stored_len = max(len(size) for size in stored)
    actual = [humanize.naturalsize(array.nbytes, binary=True) for array in arrays]
    max_actual_len = max(len(size) for size in actual)

    line = "File: {}\t{}".format(
        path, humanize.naturalsize(os.path.getsize(path), binary=True)
    )
    print(line)
    if verbosity > 0:
        print("format_version:", root.attrs["format_version"])
        prov = root.attrs["provenance"]
        print("provenance: ", end="")
        print(json.dumps(prov, indent=4, sort_keys=True))
    fmt = "{:<{}} {:<{}}\t{:<{}}\t{}"
    line = fmt.format(
        "name",
        max_name_len,
        "stored",
        max_stored_len,
        "actual",
        max_actual_len,
        "ratio",
    )
    print(line)
    for array, stored, actual in zip(arrays, stored, actual):
        ratio = 0
        if array.nbytes > 0:
            ratio = array.nbytes_stored / array.nbytes
        line = fmt.format(
            array.name,
            max_name_len,
            stored,
            max_stored_len,
            actual,
            max_actual_len,
            f"{ratio:.2f}",
        )
        print(line)
        if verbosity > 0:
            for line in str(array.info).splitlines():
                print("\t", line)



# MIT License
#
# Copyright (c) 2019 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Test cases for the command line interface for tszip.
"""
import pathlib
import sys
import tempfile
import unittest
from unittest import mock

import msprime
import numpy as np
import tskit

import tszip
import tszip.cli as cli


class TestException(Exception):
    """
    Custom exception we can throw for testing.
    """

    # We don't want pytest to use this as a class to test
    __test__ = False


def capture_output(func, *args, binary=False, **kwargs):
    """
    Runs the specified function and arguments, and returns the
    tuple (stdout, stderr) as strings.
    """
    with tempfile.TemporaryDirectory() as tmpdir:
        stdout_path = pathlib.Path(tmpdir) / "stdout"
        stderr_path = pathlib.Path(tmpdir) / "stderr"
        mode = "wb+" if binary else "w+"
        saved_stdout = sys.stdout
        saved_stderr = sys.stderr
        with open(stdout_path, mode) as stdout, open(stderr_path, mode) as stderr:
            try:
                sys.stdout = stdout
                sys.stderr = stderr
                with mock.patch("signal.signal"):
                    func(*args, **kwargs)
                stdout.seek(0)
                stderr.seek(0)
                stdout_output = stdout.read()
                stderr_output = stderr.read()
            finally:
                sys.stdout = saved_stdout
                sys.stderr = saved_stderr
    return stdout_output, stderr_output


class TestTszipArgumentParser(unittest.TestCase):
    """
    Tests for the tszip argument parser.
    """

    def test_default_values(self):
        parser = cli.tszip_cli_parser()
        infile = "tmp.trees"
        args = parser.parse_args([infile])
        self.assertEqual(args.files, [infile])
        self.assertEqual(args.keep, False)
        self.assertEqual(args.force, False)
        self.assertEqual(args.decompress, False)
        self.assertEqual(args.list, False)
        self.assertEqual(args.stdout, False)
        self.assertEqual(args.variants_only, False)
        self.assertEqual(args.suffix, ".tsz")

    def test_many_files(self):
        parser = cli.tszip_cli_parser()
        infiles = [f"{j}.trees" for j in range(1025)]
        args = parser.parse_args(infiles)
        self.assertEqual(args.files, infiles)

    def test_suffix(self):
        parser = cli.tszip_cli_parser()
        infile = "tmp.trees"
        args = parser.parse_args([infile, "-S", ".XYZ"])
        self.assertEqual(args.suffix, ".XYZ")
        args = parser.parse_args([infile, "--suffix=abx"])
        self.assertEqual(args.suffix, "abx")

    def test_decompress(self):
        parser = cli.tszip_cli_parser()
        infile = "tmp.trees.tsz"
        args = parser.parse_args([infile, "-d"])
        self.assertTrue(args.decompress)
        args = parser.parse_args([infile, "--decompress"])
        self.assertTrue(args.decompress)


class TestCli(unittest.TestCase):
    """
    Superclass of tests that run the CLI.
    """

    # Need to mock out setup_logging here or we spew logging to the console
    # in later tests.
    @mock.patch("tszip.cli.setup_logging")
    def run_tszip(self, command, mock_setup_logging):
        stdout, stderr = capture_output(cli.tszip_main, command)
        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertTrue(mock_setup_logging.called)

    @mock.patch("tszip.cli.setup_logging")
    def run_tsunzip(self, command, mock_setup_logging):
        stdout, stderr = capture_output(cli.tsunzip_main, command)
        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        self.assertTrue(mock_setup_logging.called)

    @mock.patch("tszip.cli.setup_logging")
    def run_tszip_stdout(self, command, mock_setup_logging):
        stdout, stderr = capture_output(cli.tszip_main, command, binary=True)
        self.assertEqual(stderr, b"")
        self.assertTrue(mock_setup_logging.called)
        return stdout, stderr

    @mock.patch("tszip.cli.setup_logging")
    def run_tsunzip_stdout(self, command, mock_setup_logging):
        stdout, stderr = capture_output(cli.tsunzip_main, command, binary=True)
        self.assertEqual(stderr, b"")
        self.assertTrue(mock_setup_logging.called)
        return stdout, stderr


class TestBadFiles(TestCli):
    """
    Tests that we deal with IO errors appropriately.
    """

    def test_sys_exit(self):
        # We test for cli.exit elsewhere as it's easier, but test that sys.exit
        # is called here, so we get coverage.
        with mock.patch("sys.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip(["/no/such/file"])
            mocked_exit.assert_called_once()
            args = mocked_exit.call_args[0]
            self.assertEqual(len(args), 1)
            self.assertIn("Error loading", args[0])

    def test_compress_missing(self):
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip(["/no/such/file"])
            mocked_exit.assert_called_once()
            args = mocked_exit.call_args[0]
            self.assertEqual(len(args), 1)
            self.assertTrue(args[0].startswith("Error loading"))

    def test_decompress_missing(self):
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip(["-d", "/no/such/file.tsz"])
            mocked_exit.assert_called_once()
            args = mocked_exit.call_args[0]
            self.assertEqual(len(args), 1)
            self.assertTrue(args[0].startswith("[Errno 2] No such file or directory"))

    def test_list_missing(self):
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip(["-l", "/no/such/file.tsz"])
            mocked_exit.assert_called_once()
            args = mocked_exit.call_args[0]
            self.assertEqual(len(args), 1)
            self.assertTrue(args[0].startswith("[Errno 2] No such file or directory"))


class TestCompressSemantics(TestCli):
    """
    Tests that the semantics of the CLI work as expected.
    """

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix="tszip_cli_")
        self.trees_path = pathlib.Path(self.tmpdir.name) / "msprime.trees"
        self.ts = msprime.simulate(10, mutation_rate=10, random_seed=1)
        self.ts.dump(str(self.trees_path))

    def tearDown(self):
        del self.tmpdir

    def test_simple(self):
        self.assertTrue(self.trees_path.exists())
        self.run_tszip([str(self.trees_path)])
        self.assertFalse(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".tsz")
        self.assertTrue(outpath.exists())
        ts = tszip.decompress(outpath)
        self.assertEqual(ts.tables, self.ts.tables)

    def test_suffix(self):
        self.assertTrue(self.trees_path.exists())
        self.run_tszip([str(self.trees_path), "-S", ".XYZasdf"])
        self.assertFalse(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".XYZasdf")
        self.assertTrue(outpath.exists())
        ts = tszip.decompress(outpath)
        self.assertEqual(ts.tables, self.ts.tables)

    def test_variants_only(self):
        self.assertTrue(self.trees_path.exists())
        self.run_tszip([str(self.trees_path), "--variants-only"])
        self.assertFalse(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".tsz")
        self.assertTrue(outpath.exists())
        ts = tszip.decompress(outpath)
        self.assertNotEqual(ts.tables, self.ts.tables)
        G1 = ts.genotype_matrix()
        G2 = self.ts.genotype_matrix()
        self.assertTrue(np.array_equal(G1, G2))

    def test_keep(self):
        self.assertTrue(self.trees_path.exists())
        self.run_tszip([str(self.trees_path), "--keep"])
        self.assertTrue(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".tsz")
        self.assertTrue(outpath.exists())
        ts = tszip.decompress(outpath)
        self.assertEqual(ts.tables, self.ts.tables)

    def test_overwrite(self):
        self.assertTrue(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".tsz")
        outpath.touch()
        self.assertTrue(self.trees_path.exists())
        self.run_tszip([str(self.trees_path), "--force"])
        self.assertFalse(self.trees_path.exists())
        self.assertTrue(outpath.exists())
        ts = tszip.decompress(outpath)
        self.assertEqual(ts.tables, self.ts.tables)

    def test_no_overwrite(self):
        self.assertTrue(self.trees_path.exists())
        outpath = pathlib.Path(str(self.trees_path) + ".tsz")
        outpath.touch()
        self.assertTrue(self.trees_path.exists())
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip([str(self.trees_path)])
            mocked_exit.assert_called_once_with(
                f"'{outpath}' already exists; use --force to overwrite"
            )

    def test_bad_file_format(self):
        self.assertTrue(self.trees_path.exists())
        with open(str(self.trees_path), "w") as f:
            f.write("xxx")
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip([str(self.trees_path)])
            mocked_exit.assert_called_once_with(
                f"Error loading '{self.trees_path}': File not in KAS format"
            )

    def test_compress_stdout(self):
        self.assertTrue(self.trees_path.exists())
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_tszip([str(self.trees_path)] + ["-c"])
            mocked_exit.assert_called_once_with(
                "Compressing to stdout not currently supported;"
                "Please see https://github.com/tskit-dev/tszip/issues/49"
            )


class DecompressSemanticsMixin:
    """
    Tests that the decompress semantics of the CLI work as expected.
    """

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix="tszip_cli_")
        self.trees_path = pathlib.Path(self.tmpdir.name) / "msprime.trees"
        self.ts = msprime.simulate(10, mutation_rate=10, random_seed=1)
        self.compressed_path = pathlib.Path(self.tmpdir.name) / "msprime.trees.tsz"
        tszip.compress(self.ts, self.compressed_path)
        self.trees_path1 = pathlib.Path(self.tmpdir.name) / "msprime1.trees"
        self.trees_path2 = pathlib.Path(self.tmpdir.name) / "msprime2.trees"
        self.ts1 = msprime.simulate(10, mutation_rate=10, random_seed=3)
        self.ts2 = msprime.simulate(10, mutation_rate=5, random_seed=4)
        self.compressed_path1 = pathlib.Path(self.tmpdir.name) / "msprime1.trees.tsz"
        self.compressed_path2 = pathlib.Path(self.tmpdir.name) / "msprime2.trees.tsz"
        tszip.compress(self.ts1, self.compressed_path1)
        tszip.compress(self.ts2, self.compressed_path2)

    def tearDown(self):
        del self.tmpdir

    def test_simple(self):
        self.assertTrue(self.compressed_path.exists())
        self.run_decompress([str(self.compressed_path)])
        self.assertFalse(self.compressed_path.exists())
        outpath = self.trees_path
        self.assertTrue(outpath.exists())
        ts = tskit.load(str(outpath))
        self.assertEqual(ts.tables, self.ts.tables)

    def test_suffix(self):
        suffix = ".XYGsdf"
        self.compressed_path = self.compressed_path.with_suffix(suffix)
        tszip.compress(self.ts, self.compressed_path)
        self.assertTrue(self.compressed_path.exists())
        self.run_decompress([str(self.compressed_path), "-S", suffix])
        self.assertFalse(self.compressed_path.exists())
        outpath = self.trees_path
        self.assertTrue(outpath.exists())
        ts = tskit.load(str(outpath))
        self.assertEqual(ts.tables, self.ts.tables)

    def test_keep(self):
        self.assertTrue(self.compressed_path.exists())
        self.run_decompress([str(self.compressed_path), "--keep"])
        self.assertTrue(self.compressed_path.exists())
        outpath = self.trees_path
        self.assertTrue(outpath.exists())
        ts = tskit.load(str(outpath))
        self.assertEqual(ts.tables, self.ts.tables)

    def test_keep_stdout(self):
        self.assertTrue(self.compressed_path.exists())
        self.run_decompress_stdout([str(self.compressed_path), "--stdout"])
        self.assertTrue(self.compressed_path.exists())
        self.run_decompress_stdout([str(self.compressed_path), "-c"])
        self.assertTrue(self.compressed_path.exists())

    def test_valid_stdout(self):
        tmp_file = pathlib.Path(self.tmpdir.name) / "stdout.trees"
        stdout, stderr = self.run_decompress_stdout(["-c", str(self.compressed_path)])
        with open(tmp_file, "wb+") as tmp:
            tmp.write(stdout)
        ts = tskit.load(str(tmp_file))
        self.assertEqual(ts.tables, self.ts.tables)
        self.assertTrue(self.compressed_path.exists())

    def test_valid_stdout_multiple(self):
        tmp_file = pathlib.Path(self.tmpdir.name) / "stdout.trees"
        with open(tmp_file, "wb+") as tmp:
            stdout, stderr = self.run_decompress_stdout(
                ["-c", str(self.compressed_path1), str(self.compressed_path2)]
            )
            tmp.write(stdout)
        with open(tmp_file) as out_tmp:
            ts1 = tskit.load(out_tmp)
            ts2 = tskit.load(out_tmp)
        self.assertEqual(ts1.tables, self.ts1.tables)
        self.assertEqual(ts2.tables, self.ts2.tables)
        self.assertTrue(self.compressed_path1.exists())
        self.assertTrue(self.compressed_path2.exists())

    def test_overwrite(self):
        self.assertTrue(self.compressed_path.exists())
        outpath = self.trees_path
        outpath.touch()
        self.run_decompress([str(self.compressed_path), "-f"])
        self.assertFalse(self.compressed_path.exists())
        self.assertTrue(outpath.exists())
        ts = tskit.load(str(outpath))
        self.assertEqual(ts.tables, self.ts.tables)

    def test_no_overwrite(self):
        self.assertTrue(self.compressed_path.exists())
        outpath = self.trees_path
        outpath.touch()
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_decompress([str(self.compressed_path)])
            mocked_exit.assert_called_once_with(
                f"'{outpath}' already exists; use --force to overwrite"
            )

    def test_decompress_bad_suffix(self):
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_decompress([str(self.compressed_path), "-S", "asdf"])
            mocked_exit.assert_called_once_with(
                "Compressed file must have 'asdf' suffix"
            )

    def test_bad_file_format(self):
        self.assertTrue(self.compressed_path.exists())
        with open(str(self.compressed_path), "w") as f:
            f.write("xxx")
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                self.run_decompress([str(self.compressed_path)])
            mocked_exit.assert_called_once_with(
                "Error reading '{}': File is not in tgzip format".format(
                    self.compressed_path
                )
            )


class TestDecompressSemanticsTszip(DecompressSemanticsMixin, TestCli):
    def run_decompress(self, args):
        self.run_tszip(["-d"] + args)

    def run_decompress_stdout(self, args):
        x = self.run_tszip_stdout(["-d"] + args)
        return x


class TestDecompressSemanticsTsunzip(DecompressSemanticsMixin, TestCli):
    def run_decompress(self, args):
        self.run_tsunzip(args)

    def run_decompress_stdout(self, args):
        x = self.run_tsunzip_stdout(args)
        return x


class TestList(unittest.TestCase):
    """
    Tests that the --list option works as expected.

    We don't need to mock out setup_logging here because it's not called for list.
    """

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory(prefix="tszip_cli_")
        self.trees_path = pathlib.Path(self.tmpdir.name) / "msprime.trees"
        self.ts = msprime.simulate(10, mutation_rate=10, random_seed=1)
        self.compressed_path = pathlib.Path(self.tmpdir.name) / "msprime.trees.tsz"
        tszip.compress(self.ts, self.compressed_path)

    def tearDown(self):
        del self.tmpdir

    def test_simple(self):
        stdout, stderr = capture_output(
            cli.tszip_main, ["--list", str(self.compressed_path)]
        )
        self.assertEqual(stderr, "")
        lines = stdout.splitlines()
        self.assertTrue(lines[0].startswith(f"File: {self.compressed_path}"))
        for line in lines:
            self.assertGreater(len(line), 0)

    def test_verbose(self):
        stdout, stderr = capture_output(
            cli.tszip_main, ["--list", "-v", str(self.compressed_path)]
        )
        self.assertEqual(stderr, "")
        lines = stdout.splitlines()
        self.assertTrue(lines[0].startswith(f"File: {self.compressed_path}"))
        for line in lines:
            self.assertGreater(len(line), 0)

    def test_bad_file_format(self):
        self.assertTrue(self.compressed_path.exists())
        with open(str(self.compressed_path), "w") as f:
            f.write("xxx")
        with mock.patch("tszip.cli.exit", side_effect=TestException) as mocked_exit:
            with self.assertRaises(TestException):
                cli.tszip_main([str(self.compressed_path), "-l"])
            mocked_exit.assert_called_once_with(
                "Error reading '{}': File is not in tgzip format".format(
                    self.compressed_path
                )
            )


class TestSetupLogging(unittest.TestCase):
    """
    Tests that setup logging has the desired effect.
    """

    def test_default(self):
        parser = cli.tszip_cli_parser()
        args = parser.parse_args(["afile"])
        with mock.patch("logging.basicConfig") as mocked_setup:
            cli.setup_logging(args)
            mocked_setup.assert_called_once_with(level="WARN", format=cli.log_format)

    def test_verbose(self):
        parser = cli.tszip_cli_parser()
        args = parser.parse_args(["afile", "-v"])
        with mock.patch("logging.basicConfig") as mocked_setup:
            cli.setup_logging(args)
            mocked_setup.assert_called_once_with(level="INFO", format=cli.log_format)

    def test_very_verbose(self):
        parser = cli.tszip_cli_parser()
        args = parser.parse_args(["afile", "-vv"])
        with mock.patch("logging.basicConfig") as mocked_setup:
            cli.setup_logging(args)
            mocked_setup.assert_called_once_with(level="DEBUG", format=cli.log_format)



# MIT License
#
# Copyright (c) 2019 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Command line interfaces to tszip.
"""
import argparse
import contextlib
import logging
import pathlib
import sys

import tskit

import tszip
from . import exceptions

logger = logging.getLogger(__name__)
log_format = "%(asctime)s %(levelname)s %(message)s"


def exit(message):  # noqa: A001
    """
    Exit with the specified error message, setting error status.
    """
    sys.exit(f"{sys.argv[0]}: {message}")


def setup_logging(args):
    log_level = "WARN"
    if args.verbosity > 0:
        log_level = "INFO"
    if args.verbosity > 1:
        log_level = "DEBUG"
    logging.basicConfig(level=log_level, format=log_format)


def tszip_cli_parser():
    parser = argparse.ArgumentParser(
        description="Compress/decompress tskit trees files."
    )
    parser.add_argument(
        "-V", "--version", action="version", version=f"%(prog)s {tszip.__version__}"
    )
    parser.add_argument(
        "-v", "--verbosity", action="count", default=0, help="Increase the verbosity"
    )
    parser.add_argument("files", nargs="+", help="The files to compress/decompress.")
    parser.add_argument(
        "--variants-only",
        action="store_true",
        help=(
            "Lossy compression; throws out information not needed to "
            "represent variants"
        ),
    )
    parser.add_argument(
        "-S", "--suffix", default=".tsz", help="Use suffix SUFFIX on compressed files"
    )
    parser.add_argument(
        "-k", "--keep", action="store_true", help="Keep (don't delete) input files"
    )
    parser.add_argument(
        "-f", "--force", action="store_true", help="Force overwrite of output file"
    )
    parser.add_argument("-c", "--stdout", action="store_true", help="Write to stdout")
    group = parser.add_mutually_exclusive_group()
    group.add_argument("-d", "--decompress", action="store_true", help="Decompress")
    group.add_argument(
        "-l", "--list", action="store_true", help="List contents of the file"
    )
    return parser


def remove_input(infile, args):
    if not args.keep:
        logger.info(f"Removing {infile}")
        infile.unlink()


def check_output(outfile, args):
    if outfile.exists():
        if not args.force:
            exit(f"'{outfile}' already exists; use --force to overwrite")


def run_compress(args):
    if args.stdout:
        exit(
            "Compressing to stdout not currently supported;"
            "Please see https://github.com/tskit-dev/tszip/issues/49"
        )
    setup_logging(args)
    for file_arg in args.files:
        logger.info(f"Compressing {file_arg}")
        try:
            ts = tskit.load(file_arg)
        except (FileNotFoundError, tskit.FileFormatError) as ffe:
            exit(f"Error loading '{file_arg}': {ffe}")
        logger.debug("Loaded tree sequence")
        infile = pathlib.Path(file_arg)
        outfile = pathlib.Path(file_arg + args.suffix)
        check_output(outfile, args)
        tszip.compress(ts, outfile, variants_only=args.variants_only)
        remove_input(infile, args)


@contextlib.contextmanager
def check_load_errors(file_arg):
    try:
        yield
    except OSError as ose:
        exit(str(ose))
    except exceptions.FileFormatError as ffe:
        exit(f"Error reading '{file_arg}': {ffe}")


def run_decompress(args):
    setup_logging(args)
    for file_arg in args.files:
        logger.info(f"Decompressing {file_arg}")
        if not file_arg.endswith(args.suffix):
            exit(f"Compressed file must have '{args.suffix}' suffix")
        infile = pathlib.Path(file_arg)
        if args.stdout:
            args.keep = True
            outfile = sys.stdout
        else:
            outfile = pathlib.Path(file_arg[: -len(args.suffix)])
            check_output(outfile, args)
        with check_load_errors(file_arg):
            ts = tszip.decompress(file_arg)
        ts.dump(outfile)
        logger.info(f"Wrote {outfile}")
        remove_input(infile, args)


def run_list(args):
    for file_arg in args.files:
        with check_load_errors(file_arg):
            tszip.print_summary(file_arg, args.verbosity)


def main(args):
    if args.decompress:
        run_decompress(args)
    elif args.list:
        run_list(args)
    else:
        run_compress(args)


def tszip_main(arg_list=None):
    parser = tszip_cli_parser()
    args = parser.parse_args(arg_list)
    main(args)


def tsunzip_main(arg_list=None):
    parser = tszip_cli_parser()
    args = parser.parse_args(arg_list)
    args.decompress = True
    main(args)



# MIT License
#
# Copyright (c) 2021 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Tests for the basic compression functionality.
"""
import pathlib
import tempfile
import unittest

import msprime
import numpy as np
import tskit
import zarr

import tszip
import tszip.compression as compression
import tszip.exceptions as exceptions
import tszip.provenance as provenance


class TestMinimalDtype(unittest.TestCase):
    """
    Test that we compute the minimal dtype for data correctly.
    """

    def verify(self, dtype, values):
        for v in values:
            min_dtype = compression.minimal_dtype(np.array([v]))
            self.assertEqual(min_dtype, np.dtype(dtype))

    def test_int8(self):
        self.verify(np.int8, np.array([0, -1, 1, 127, -127], dtype=np.int64))

    def test_uint8(self):
        self.verify(np.uint8, np.array([0, 1, 127, 255], dtype=np.uint64))

    def test_int16(self):
        self.verify(
            np.int16,
            np.array(
                [2 ** 15 - 1, -(2 ** 15) + 1, 2 ** 7 + 1, -(2 ** 7) - 1], dtype=np.int64
            ),
        )

    def test_uint16(self):
        self.verify(np.uint16, np.array([256, 2 ** 16 - 1], dtype=np.uint64))

    def test_int32(self):
        self.verify(
            np.int32,
            np.array(
                [2 ** 31 - 1, -(2 ** 31) + 1, 2 ** 15 + 1, -(2 ** 15) - 1],
                dtype=np.int64,
            ),
        )

    def test_uint32(self):
        self.verify(np.uint32, np.array([2 ** 16 + 1, 2 ** 32 - 1], dtype=np.uint64))

    def test_int64(self):
        self.verify(
            np.int64,
            np.array(
                [2 ** 63 - 1, -(2 ** 63) + 1, 2 ** 31 + 1, -(2 ** 31) - 1],
                dtype=np.int64,
            ),
        )

    def test_uint64(self):
        self.verify(np.uint64, np.array([2 ** 32 + 1, 2 ** 64 - 1], dtype=np.uint64))

    def test_float32(self):
        self.verify(np.float32, np.array([0.1, 1e-3], dtype=np.float32))

    def test_float64(self):
        self.verify(np.float64, np.array([0.1, 1e-3], dtype=np.float64))

    def test_empty(self):
        for dtype in map(np.dtype, [np.float64, np.int32, np.uint8]):
            min_dtype = compression.minimal_dtype(np.array([], dtype=dtype))
            self.assertEqual(min_dtype, dtype)


class RoundTripMixin:
    """
    Set of example tree sequences that we should be able to round trip.
    """

    def test_small_msprime_no_recomb(self):
        ts = msprime.simulate(10, mutation_rate=2, random_seed=2)
        self.assertGreater(ts.num_sites, 2)
        self.verify(ts)

    def test_small_msprime_recomb(self):
        ts = msprime.simulate(10, recombination_rate=2, mutation_rate=2, random_seed=2)
        self.assertGreater(ts.num_sites, 2)
        self.assertGreater(ts.num_trees, 2)
        self.verify(ts)

    def test_small_msprime_migration(self):
        ts = msprime.simulate(
            population_configurations=[
                msprime.PopulationConfiguration(10),
                msprime.PopulationConfiguration(10),
            ],
            migration_matrix=[[0, 1], [1, 0]],
            record_migrations=True,
            recombination_rate=2,
            mutation_rate=2,
            random_seed=2,
        )
        self.assertGreater(ts.num_sites, 2)
        self.assertGreater(ts.num_migrations, 1)
        self.assertGreater(ts.num_trees, 2)
        self.verify(ts)

    def test_small_msprime_top_level_metadata(self):
        ts = msprime.simulate(10, recombination_rate=2, mutation_rate=2, random_seed=2)
        self.assertGreater(ts.num_sites, 2)
        self.assertGreater(ts.num_trees, 2)
        tables = ts.dump_tables()
        top_level_schema = {
            "codec": "json",
            "properties": {"my_int": {"type": "integer"}},
        }
        tables.metadata_schema = tskit.MetadataSchema(top_level_schema)
        tables.metadata = {"my_int": 1234}
        self.verify(tables.tree_sequence())

    def test_small_msprime_individuals_metadata(self):
        ts = msprime.simulate(10, recombination_rate=1, mutation_rate=2, random_seed=2)
        self.assertGreater(ts.num_sites, 2)
        self.assertGreater(ts.num_trees, 2)
        tables = ts.dump_tables()
        tables.nodes.clear()
        for j, node in enumerate(ts.nodes()):
            tables.individuals.add_row(
                flags=j, location=[j] * j, parents=[j - 1] * j, metadata=b"x" * j
            )
            tables.nodes.add_row(
                flags=node.flags,
                population=node.population,
                individual=j,
                time=node.time,
                metadata=b"y" * j,
            )
        tables.populations.clear()
        tables.populations.add_row(metadata=b"X" * 1024)
        tables.sort()
        self.verify(tables.tree_sequence())

    def test_small_msprime_complex_mutations(self):
        ts = msprime.simulate(
            10, recombination_rate=0.1, mutation_rate=0.2, random_seed=2, length=10
        )
        tables = ts.dump_tables()
        tables.sites.clear()
        tables.mutations.clear()
        for j, site in enumerate(ts.sites()):
            tables.sites.add_row(
                position=site.position, ancestral_state=j * "A", metadata=j * b"x"
            )
            for mutation in site.mutations:
                tables.mutations.add_row(
                    mutation.site,
                    node=mutation.node,
                    derived_state=(j + 1) * "T",
                    metadata=j * b"y",
                )
        self.verify(tables.tree_sequence())

    def test_mutation_parent_example(self):
        tables = tskit.TableCollection(1)
        tables.nodes.add_row(flags=tskit.NODE_IS_SAMPLE, time=0)
        tables.nodes.add_row(flags=tskit.NODE_IS_SAMPLE, time=0)
        tables.sites.add_row(position=0, ancestral_state="A")
        tables.mutations.add_row(site=0, node=0, derived_state="T")
        tables.mutations.add_row(site=0, node=0, parent=0, derived_state="A")
        self.verify(tables.tree_sequence())

    def test_all_fields(self):
        demography = msprime.Demography()
        demography.add_population(name="A", initial_size=10_000)
        demography.add_population(name="B", initial_size=5_000)
        demography.add_population(name="C", initial_size=1_000)
        demography.add_population_split(time=1000, derived=["A", "B"], ancestral="C")
        ts = msprime.sim_ancestry(
            samples={"A": 1, "B": 1},
            demography=demography,
            random_seed=42,
            record_migrations=True,
        )
        ts = msprime.sim_mutations(ts, rate=1, random_seed=42)
        tables = ts.dump_tables()
        for name, table in tables.name_map.items():
            if name not in ["provenances", "edges"]:
                table.metadata_schema = tskit.MetadataSchema({"codec": "json"})
                metadatas = [f'{{"foo":"n_{name}_{u}"}}' for u in range(len(table))]
                metadata, metadata_offset = tskit.pack_strings(metadatas)
                table.set_columns(
                    **{
                        **table.asdict(),
                        "metadata": metadata,
                        "metadata_offset": metadata_offset,
                    }
                )
        tables.metadata_schema = tskit.MetadataSchema({"codec": "json"})
        tables.metadata = "Test metadata"
        self.verify(tables.tree_sequence())


class TestGenotypeRoundTrip(unittest.TestCase, RoundTripMixin):
    """
    Tests that we can correctly roundtrip genotype data losslessly.
    """

    def verify(self, ts):
        if ts.num_migrations > 0:
            raise unittest.SkipTest("Migrations not supported")
        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir) / "treeseq.tsz"
            tszip.compress(ts, path, variants_only=True)
            other_ts = tszip.decompress(path)
        self.assertEqual(ts.num_sites, other_ts.num_sites)
        for var1, var2 in zip(ts.variants(), other_ts.variants()):
            self.assertTrue(np.array_equal(var1.genotypes, var2.genotypes))
            self.assertEqual(var1.site.position, var2.site.position)
            self.assertEqual(var1.alleles, var2.alleles)
        # Populations, individuals and sites should be untouched if there are no
        # unreachable individuals.
        t1 = ts.tables
        t2 = other_ts.tables
        self.assertEqual(t1.sequence_length, t2.sequence_length)
        self.assertEqual(t1.populations, t2.populations)
        self.assertEqual(t1.individuals, t2.individuals)
        self.assertEqual(t1.sites, t2.sites)
        # We should be adding an extra provenance record in here due to simplify.
        self.assertEqual(len(t1.provenances), len(t2.provenances) - 1)


class TestExactRoundTrip(unittest.TestCase, RoundTripMixin):
    def verify(self, ts):
        with tempfile.TemporaryDirectory() as tmpdir:
            path = pathlib.Path(tmpdir) / "treeseq.tsz"
            tszip.compress(ts, path)
            other_ts = tszip.decompress(path)
        self.assertEqual(ts.tables, other_ts.tables)


class TestMetadata(unittest.TestCase):
    """
    Tests that we correctly write the format information to the file and
    that we read it also.
    """

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.path = pathlib.Path(self.tmpdir.name) / "treeseq.tsz"

    def tearDown(self):
        del self.tmpdir

    def test_format_written(self):
        ts = msprime.simulate(10, random_seed=1)
        tszip.compress(ts, self.path)
        with zarr.ZipStore(str(self.path), mode="r") as store:
            root = zarr.group(store=store)
            self.assertEqual(root.attrs["format_name"], compression.FORMAT_NAME)
            self.assertEqual(root.attrs["format_version"], compression.FORMAT_VERSION)

    def test_provenance(self):
        ts = msprime.simulate(10, random_seed=1)
        for variants_only in [True, False]:
            tszip.compress(ts, self.path, variants_only=variants_only)
            with zarr.ZipStore(str(self.path), mode="r") as store:
                root = zarr.group(store=store)
                self.assertEqual(
                    root.attrs["provenance"],
                    provenance.get_provenance_dict({"variants_only": variants_only}),
                )

    def write_file(self, attrs, path):
        with zarr.ZipStore(str(path), mode="w") as store:
            root = zarr.group(store=store)
            root.attrs.update(attrs)

    def test_missing_format_keys(self):
        values = [{}, {"format_name": ""}, {"format_version": ""}]
        for attrs in values:
            self.write_file(attrs, self.path)
            with self.assertRaises(exceptions.FileFormatError):
                tszip.decompress(self.path)
            with self.assertRaises(exceptions.FileFormatError):
                tszip.print_summary(self.path)

    def test_bad_format_name(self):
        for bad_name in ["", "xyz", [1234]]:
            self.write_file(
                {"format_name": bad_name, "format_version": [1, 0]}, self.path
            )
            with self.assertRaises(exceptions.FileFormatError):
                tszip.decompress(self.path)
            with self.assertRaises(exceptions.FileFormatError):
                tszip.print_summary(self.path)

    def test_format_too_old(self):
        self.write_file({"format_name": "tszip", "format_version": [0, 0]}, self.path)
        with self.assertRaises(exceptions.FileFormatError):
            tszip.decompress(self.path)
        with self.assertRaises(exceptions.FileFormatError):
            tszip.print_summary(self.path)

    def test_format_too_new(self):
        self.write_file({"format_name": "tszip", "format_version": [2, 0]}, self.path)
        with self.assertRaises(exceptions.FileFormatError):
            tszip.decompress(self.path)
        with self.assertRaises(exceptions.FileFormatError):
            tszip.print_summary(self.path)

    def test_wrong_format(self):
        for contents in ["", "1234", "X" * 1024]:
            with open(str(self.path), "w") as f:
                f.write(contents)
            with self.assertRaises(exceptions.FileFormatError):
                tszip.decompress(self.path)


class TestFileErrors(unittest.TestCase):
    """
    Tests that we correctly write the format information to the file and
    that we read it also.
    """

    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.path = pathlib.Path(self.tmpdir.name) / "treeseq.tsz"

    def tearDown(self):
        del self.tmpdir

    def test_missing_file(self):
        path = "/no/such/file"
        with self.assertRaises(FileNotFoundError):
            tszip.decompress(path)

    def test_load_dir(self):
        with self.assertRaises(OSError):
            tszip.decompress(self.path.parent)

    def test_save_dir(self):
        ts = msprime.simulate(10, random_seed=1)
        with self.assertRaises(OSError):
            tszip.compress(ts, self.path.parent)



# MIT License
#
# Copyright (c) 2019 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Versions of important dependencies and environment.
"""
import platform

import numcodecs
import tskit
import zarr


__version__ = "undefined"
try:
    from . import _version

    __version__ = _version.version
except ImportError:  # pragma: nocover
    pass


def get_environment():
    """
    Returns a dictionary encoding provenance information
    """
    env = {
        "os": {
            "system": platform.system(),
            "node": platform.node(),
            "release": platform.release(),
            "version": platform.version(),
            "machine": platform.machine(),
        },
        "python": {
            "implementation": platform.python_implementation(),
            "version": platform.python_version(),
        },
    }
    libs = {
        "tskit": {"version": tskit.__version__},
        "numcodecs": {"version": numcodecs.__version__},
        "zarr": {"version": zarr.__version__},
    }
    env["libraries"] = libs
    return env


def get_provenance_dict(parameters):
    """
    Returns a dictionary encoding an execution of tszip conforming to the
    tskit provenance schema.
    """
    document = {
        "schema_version": "1.0.0",
        "software": {"name": "tszip", "version": __version__},
        "parameters": parameters,
        "environment": get_environment(),
    }
    return document



# MIT License
#
# Copyright (c) 2019 Tskit Developers
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Tests for the basic provenance recording.
"""
import unittest

import numcodecs
import tskit
import zarr

import tszip
from tszip import provenance


class TestProvenance(unittest.TestCase):
    """
    Test basic provenance properties.
    """

    def test_schema_validates(self):
        for params in [{}, {"a": "a"}, {"a": {"a": 1}}]:
            d = provenance.get_provenance_dict(params)
            tskit.validate_provenance(d)

    def test_environment(self):
        # Basic environment should be the same as tskit.
        d_tszip = provenance.get_provenance_dict({})
        d_tskit = tskit.provenance.get_provenance_dict()
        self.assertEqual(d_tszip["environment"]["os"], d_tskit["environment"]["os"])
        self.assertEqual(
            d_tszip["environment"]["python"], d_tskit["environment"]["python"]
        )

    def test_libraries(self):
        libs = provenance.get_provenance_dict({})["environment"]["libraries"]
        self.assertEqual(libs["tskit"]["version"], tskit.__version__)
        self.assertEqual(libs["zarr"]["version"], zarr.__version__)
        self.assertEqual(libs["numcodecs"]["version"], numcodecs.__version__)

    def test_sofware(self):
        software = provenance.get_provenance_dict({})["software"]
        self.assertEqual(software, {"name": "tszip", "version": tszip.__version__})

    def test_parameters(self):
        for params in [{}, {"a": "a"}, {"a": {"a": 1}}]:
            d = provenance.get_provenance_dict(params)["parameters"]
            self.assertEqual(d, params)
